﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Forms;
using System.IO;
using System.Collections;

namespace TagCloud
{
    public partial class TagCloud : Form
    {
        public class TagData
        {
            private int Vol;
            private string TP;
            private int sz;
            private int pc;
            public int Volume
            {
                get { return Vol; }
                set { Vol = value; }
            }
            public string TagPhrase
            {
                get { return TP; }
                set { TP = value; }
            }
            public int Size
            {
                get { return sz; }
                set { sz = value; }
            }
            public int PixelCount
            {
                get { return pc; }
                set { pc = value; }
            }
        }

        Random rand = new Random();    //1 for no overlap; 2 for overlap
        private Graphics imgGraphics;
        private Image img;
        TagData[] arrData;
        int[] genScores;
        Rectangle[] TagLocations;
        Rectangle[][] GenLocations = new Rectangle[1500][];
        int totalchromosomes = 100;
        int total_gens = 1;
        string wordfile = "tweet_data.txt";
        int tagphrase_total = 30;       //number of tagphrases in document
        int maxX = 800;     //find a way to dynamically size x and y based on form
        int maxY = 650;
        int score = 0;
        int leftmost = 800;          //for proximity calculation - start large
        int topmost = 650;
        int rightmost, botmost = 0;           //start small and increase
        Color c = Color.FromArgb(1, 0, 0);
        SolidBrush blackbrush = new SolidBrush(Color.Black);
        SolidBrush whitebrush = new SolidBrush(Color.White);
        int pixelCT = 0;
        int totalPixels = 0;
        int offset = 0;
        Color pixcolor = Color.White;       //initializing w/value - white doesn't matter
        int tpHalf = 0;

        public TagCloud()
        {
            InitializeComponent();
            img = new Bitmap(maxX, maxY);     //for rendering function as opposed to box
            imgGraphics = Graphics.FromImage(img);
            //FIRST: BRING THE DATA IN
            //Open the file
            if (!File.Exists(wordfile))
                MessageBox.Show("File Not Found");
            string[] lines = File.ReadAllLines(wordfile);
            char[] delims = { '~' };
            string[] tstring;
            //create an array of of TagData elements: name and volume
            TagData[] TD = new TagData[tagphrase_total];     //make larger?
            Rectangle[] TL = new Rectangle[tagphrase_total];
            Rectangle[] TR = new Rectangle[tagphrase_total];
            for (int i = 0; i < tagphrase_total; i++)
            {
                TD[i] = new TagData();
                TR[i] = new Rectangle();
            }
            arrData = TD;
            TagLocations = TR;

            //populate the array with the data from the file
            for (int n = 0; n < tagphrase_total; n++)
            {
                tstring = lines[n].Split(delims);
                arrData[n].TagPhrase = tstring[0];
                arrData[n].Volume = Convert.ToInt32(tstring[1]);
                arrData[n].Size = 20;   //default size; actual size to be determined later
            }

            //SECOND: Determine font size & pixel counts for each tag phrase
            DetermineFontSize();

            tpHalf = tagphrase_total / 2;
            //PRE-PROCESSING DONE.  

            //initialize scores array
            int[] TG = new int[1500];
            for (int i = 0; i < 1500; i++)
            {
                TG[i] = 0;
            }
            genScores = TG;
            //initialize genetic algorithm locations storage array
            for (int i = 0; i < 1500; i++)
            {
                GenLocations[i] = new Rectangle[tagphrase_total];
            }

            //GENETIC ALGORITHM PART 
            //1. Randomly place the tagphrases & get the score for each configuration
            //  1. a) Store tagphrase locations in jagged array - GenLocations
            //  1. b) Store scores for each iteration in scores array
            //2. Keep good ones; throw out bad ones
            //  2. a) loop through array of scores
            //  2. b) index where scores go negative (or barely positive - as in bad?)
            //3. "Mate" the existing options - Layout A & B
            //  3. a) combine 1st half of A and last half of B
            //  3. b) combine 1st half of B and last half of A
            //  3. c) store locations/scores starting where negative started
            //  3. d) bubble sort new combinations
            //REPEAT STEPS 2 and 3

            for (int i = 0; i < totalchromosomes; i++)        //step 1 loop
            {
                //clear "scratch paper" bitmap
                Rectangle R = new Rectangle(0, 0, maxX, maxY);
                imgGraphics.FillRectangle(whitebrush, R);   //clear screen
                leftmost = 800;          //for proximity calculation - start large
                topmost = 650;
                rightmost = 0;
                botmost = 0;
                InitialScreenPlacementRender();                    //1.a
                                                            //store tagphrase locations
                for (int j = 0; j < tagphrase_total; j++)
                {
                    GenLocations[i][j] = TagLocations[j];   //1.a
                }
                //check overlap and check proximity
                score = CheckOverlap() + CheckProximity();
                genScores[i] = score;                       //1.b
            }
            BubbleSort(totalchromosomes);

            while (genScores[0] < 4500 && total_gens < 100) //steps 2 and 3; mate many
            {                                           //generations to get optimization
                int tscore = 3001;
                int gi1 = 0;
                while (tscore >= 3000)                          //2.b
                {
                    tscore = genScores[gi1];
                    gi1++;
                }

                //Mating time! - Step 3
                gi1 = MateMe(gi1);
                //check new scores
                for (int p = 0; p < gi1; p++)
                {
                    leftmost = 800;          //for proximity calculation - start large
                    topmost = 650;
                    rightmost = 0;
                    botmost = 0;
                    ScreenPlacementRender(p);       //draw valid locations to scratch for overlap
                    //check overlap and check proximity
                    score = CheckOverlap() + CheckProximity();
                    genScores[p] = score;                       //1.b
                }

                //getting lots of clones - with bubble sort; if I have 20 arrays with
                //the same score at the top, they will stay in the same position, producing
                //identical children, and I end up with a lot of redundancy.
                //So I will allow a certain amount of expansion, but every so often, I need
                //to kill some chromosomes and randomize the population.
                //If there are too many, I purge

                //shuffle first...
                Randomize_Population(gi1);
                //then sort to make sure the best ones still survive...
                BubbleSort(gi1);

                if (gi1 > 500)
                {
                    //then kill bottom chunk
                    int ShallowEndOfGenePool = 300;
                    for (int i = ShallowEndOfGenePool; i < gi1; i++)
                    {
                        genScores[i] = 0;                   //clear out score
                        GenLocations[i] = new Rectangle[tagphrase_total]; //clear out locations
                    }
                    gi1 = ShallowEndOfGenePool;      //reset value for gi1
                }

                total_gens++; //let's see how many generations it takes to get a good one
            }

            FinalRender();      //render the best tag cloud to the screen 
                                //with multiple colors for ease of reading
        }

        /****************** FUNCTIONS ******************/
        public void DetermineFontSize()
        {
            //Tweet volume for the most popular tweets runs in the tens of thousands
            //if not hundreds of thousands. To create numbers that work more easily 
            //with font sizes and algorithms,  I'm fitting them into ranges, so they are
            //something smaller and more managable.  Font sizes are not an exact 1:1
            //ratio to tweet volume; they are a general reflection of popularity, so 
            //if some information is lost in the fitting, it is not a big loss.

            Point zero = new Point(0, 0);

            for (int i = 0; i < tagphrase_total; i++)
            {
                if (arrData[i].Volume > 410000)
                    arrData[i].Size = 65;
                else if (arrData[i].Volume <= 410000 && arrData[i].Volume > 380000)
                    arrData[i].Size = 60;
                else if (arrData[i].Volume <= 380000 && arrData[i].Volume > 350000)
                    arrData[i].Size = 56;
                else if (arrData[i].Volume <= 350000 && arrData[i].Volume > 320000)
                    arrData[i].Size = 52;
                else if (arrData[i].Volume <= 320000 && arrData[i].Volume > 290000)
                    arrData[i].Size = 48;
                else if (arrData[i].Volume <= 290000 && arrData[i].Volume > 260000)
                    arrData[i].Size = 44;
                else if (arrData[i].Volume <= 260000 && arrData[i].Volume > 230000)
                    arrData[i].Size = 40;
                else if (arrData[i].Volume <= 230000 && arrData[i].Volume > 200000)
                    arrData[i].Size = 36;
                else if (arrData[i].Volume <= 200000 && arrData[i].Volume > 160000)
                    arrData[i].Size = 32;
                else if (arrData[i].Volume <= 160000 && arrData[i].Volume > 130000)
                    arrData[i].Size = 28;
                else if (arrData[i].Volume <= 130000 && arrData[i].Volume > 100000)
                    arrData[i].Size = 24;
                else if (arrData[i].Volume <= 100000 && arrData[i].Volume > 75000)
                    arrData[i].Size = 20;
                else if (arrData[i].Volume <= 75000 && arrData[i].Volume > 45000)
                    arrData[i].Size = 16;
                else if (arrData[i].Volume <= 45000 && arrData[i].Volume > 25000)
                    arrData[i].Size = 13;
                else if (arrData[i].Volume <= 25000 && arrData[i].Volume > 10000)
                    arrData[i].Size = 10;
                else                //anything less than 10,000 volume
                    arrData[i].Size = 7;

                //get pixels for each tag phrase
                Font tfont = new Font("Arial", arrData[i].Size, FontStyle.Regular);
                Size ts1 = TextRenderer.MeasureText(arrData[i].TagPhrase, tfont);
                Size ts2 = TextRenderer.MeasureText(arrData[i].TagPhrase, tfont, ts1, 
                    TextFormatFlags.GlyphOverhangPadding);    //handles the descent
                int wid = (ts2.Width);
                int hgt = (ts2.Height);
                Rectangle R = new Rectangle(zero, ts2);
                TextRenderer.DrawText(imgGraphics, arrData[i].TagPhrase, tfont,
                    zero, c);                       //barely not black text
                Bitmap tempBMP = new Bitmap(img);
                pixelCT = 0;    //reset value
                for (int h = 0; h <= hgt; h++)       //move through the rectangle
                {                                   //looking for pixels that match
                    for (int w = 0; w <= wid; w++)    //the color of the text
                    {
                        pixcolor = tempBMP.GetPixel(w, h);  
                        if (pixcolor == c)          //if the colors match, increase count
                            pixelCT++;
                    }
                }
                arrData[i].PixelCount = pixelCT;    //store in data array
                totalPixels += pixelCT;             //add new pixels to total pixels
                imgGraphics.FillRectangle(whitebrush, R);   //clear out any existing pixels
            }         
        }

        private void InitialScreenPlacementRender()
        {
            Rectangle[] tr = new Rectangle[tagphrase_total];
            for (int i = 0; i < tagphrase_total; i++)
            {
                tr[i] = new Rectangle();
            }
            for (int i = 0; i<tagphrase_total; i++)
            {
                Font tfont = new Font("Arial", arrData[i].Size, FontStyle.Regular);
                    //find out how big the text will be
                Size tsize = TextRenderer.MeasureText(arrData[i].TagPhrase, tfont);
                    //find a random spot to draw it
                int x = maxX - tsize.Width;
                int y = maxY - tsize.Height;
                int xplace = rand.Next(x);   //prevents from going off screen
                int yplace = rand.Next(y);   //by subtracting dimensions
                Point rendLoc = new Point(xplace, yplace);
                
                Rectangle Rect = new Rectangle(rendLoc, tsize);
                TextRenderer.DrawText(imgGraphics, arrData[i].TagPhrase, tfont, 
                    rendLoc, c);
                tr[i] = Rect;   //to track size and location of each rectangle
                TagLocations[i] = new Rectangle();
                TagLocations[i] = Rect;

                //keep track of edges of cloud for optimization
                if (Rect.Right > rightmost)
                    rightmost = Rect.Right;
                if (Rect.Left < leftmost)
                    leftmost = Rect.Left;
                if (Rect.Top < topmost)
                    topmost = Rect.Top;
                if (Rect.Bottom > botmost)
                    botmost = Rect.Bottom;
            }
        }

        private void ScreenPlacementRender(int ind)
        {
            Rectangle R = new Rectangle(0, 0, maxX, maxY);
            imgGraphics.FillRectangle(whitebrush, R);   //clear screen
            for (int i = 0; i < tagphrase_total; i++)
            {
                Font tfont = new Font("Arial", arrData[i].Size, FontStyle.Regular);
                Point rendLoc = new Point(GenLocations[ind][i].Left, GenLocations[ind][i].Top);
                TextRenderer.DrawText(imgGraphics, arrData[i].TagPhrase, tfont,
                    rendLoc, c);
                if (GenLocations[ind][i].Right > rightmost)
                    rightmost = GenLocations[ind][i].Right;
                if (GenLocations[ind][i].Left < leftmost)
                    leftmost = GenLocations[ind][i].Left;
                if (GenLocations[ind][i].Top < topmost)
                    topmost = GenLocations[ind][i].Top;
                if (GenLocations[ind][i].Bottom > botmost)
                    botmost = GenLocations[ind][i].Bottom;
            }
        }

        private void FinalRender()
        {
            MessageBox.Show("best score: " + genScores[0].ToString() + 
                " total generations: " + total_gens.ToString());
            Rectangle R = new Rectangle(0, 0, maxX, maxY);
            imgGraphics.FillRectangle(whitebrush, R);   //clear screen
            for (int i = 0; i < tagphrase_total; i++)
            {
                Font tfont = new Font("Arial", arrData[i].Size, FontStyle.Regular);
                Point rendLoc = new Point(GenLocations[0][i].Left, GenLocations[0][i].Top);                           
                if ((i % 3) == 0)           //rotating through color options
                    c = Color.Black;        //color either Black, Red or Blue, depending
                else if ((i % 3) == 1)      //on mod 3 value
                    c = Color.Red;
                else
                    c = Color.Blue;
                TextRenderer.DrawText(imgGraphics, arrData[i].TagPhrase, tfont,
                    rendLoc, c);
            }
        }

        private int CheckOverlap()
        {
            int rtnvalue = 0;
            Bitmap tempBMP = new Bitmap(img);
            pixelCT = 0;    //reset value
            for (int h = 0; h < maxY; h++)
            {
                for (int w = 0; w < maxX; w++)
                {
                    pixcolor = tempBMP.GetPixel(w, h);
                    if (pixcolor == c)          //if the colors match, increase count
                        pixelCT++;
                }
            }
                                        //if font colored pixels in whole image < 
            if (pixelCT < (totalPixels-35))      //what total pixel count should be, then
                rtnvalue = -8000;          //something overlaps something else
            else
                rtnvalue = 3000;            //no overlap; return a positive score
            return rtnvalue;
        }

        private int CheckProximity()
        {
            int rv = 0;
                    //how far apart is everything?
            int LRDiff = rightmost - leftmost;      //gives width
            int TBDiff = botmost - topmost;         //gives height

            if (LRDiff > 0 && LRDiff < 200)     //good, close cloud
                rv = 2000;                      //good return value
            else if (LRDiff >= 200 && LRDiff < 300) //a little looser
                rv = 1500;
            else if (LRDiff >= 300 && LRDiff < 400) //medium cloud
                rv = 1000;
            else if (LRDiff >= 400 && LRDiff < 600)     //getting big
                rv = 500;
            else if (LRDiff >= 600)         //800 max width; 600-800 quite wide
                rv = 0;

            if (TBDiff > 0 && TBDiff < 150)     //not too tall
                rv += 2000;
            else if (TBDiff >= 150 && TBDiff < 250)
                rv += 1500;
            else if (TBDiff >= 250 && TBDiff < 350)
                rv += 1000;
            else if (TBDiff >= 350 && TBDiff < 450)
                rv += 500;
            else if (LRDiff >= 450)             //way too tall
                rv += 0;

            return rv;
        }

        private void BubbleSort(int size)
        {
            int temp = 0;
            Rectangle[] temp2;
            for (int i = 0; i < size; i++)
            {
                for (int j = 0; j < size-1; j++)
                {
                    if (genScores[j] < genScores[j+1])
                    {
                        temp = genScores[j + 1];
                        genScores[j + 1] = genScores[j];
                        genScores[j] = temp;
                        temp2 = GenLocations[j + 1];
                        GenLocations[j + 1] = GenLocations[j];
                        GenLocations[j] = temp2;
                    }
                }
            }
        }

        /*** MATE ME***/
        //3. a/c  mating 1st half A + 2nd half B
        //Also, check each location if it's an outlier; if so generate new location
        //instead of slowly moving things in (like in SnugUp), generate an 
        //entirely new random location for the outlier word
        //If I had the processing power to do both SnugUp and random locations, I would
        //The more variation in the chromosomes, the better... at least in concept.
        private int MateMe(int gi2)
        {
            offset = gi2;
            Rectangle Pass;
            for (int i = 0; i < gi2; i += 2)         //mating two at a time
            {
                for (int j = 0; j < tpHalf; j++)
                {
                    //copy content, then mutate
                    //check for outlier before saving
                    Pass = GenLocations[i][j];
                    if (Pass.Top < 200 || Pass.Bottom > 400 || Pass.Left < 100 || Pass.Right > 600)
                    {       //prevents from going off screen by subtracting dimensions
//                        GenLocations[offset + 1][j] = SnugUp(Pass);  //snugging up and randomizing
                        Pass.X = rand.Next(maxX - (Pass.Width));
                        Pass.Y = rand.Next(maxY - (Pass.Height));
                    }
                    GenLocations[offset][j] = Pass;
                }
                for (int k = tpHalf; k < tagphrase_total; k++)
                {
                    //mutate and store
                    Pass = GenLocations[i + 1][k];
                    if (Pass.Top < 200 || Pass.Bottom > 400 || Pass.Left < 100 || Pass.Right > 600)
                    {
//                        GenLocations[offset + 1][k] = SnugUp(Pass);
                        Pass.X = rand.Next(maxX - (Pass.Width));
                        Pass.Y = rand.Next(maxY - (Pass.Height));
                    }
                    GenLocations[offset][k] = Pass;
                }
                offset = offset+2;
                //3. b/c  mating 2nd half of A + 1st half of B in next available spot
                for (int j = 0; j < tpHalf; j++)
                {
                    Pass = GenLocations[i + 1][j];
                    if (Pass.Top < 200 || Pass.Bottom > 400 || Pass.Left < 100 || Pass.Right > 600)
                    {
//                       GenLocations[offset + 1][j] = SnugUp(Pass);
                        Pass.X = rand.Next(maxX - (Pass.Width));
                        Pass.Y = rand.Next(maxY - (Pass.Height));
                    }
                    GenLocations[offset][j] = Pass;
                }
                for (int k = tpHalf; k < tagphrase_total; k++)
                {
                    Pass = GenLocations[i][k];
                    if (Pass.Top < 100 || Pass.Bottom > 400 || Pass.Left < 100 || Pass.Right > 600)
                    {
//                        GenLocations[offset + 1][k] = SnugUp(Pass);
                        Pass.X = rand.Next(maxX - (Pass.Width));
                        Pass.Y = rand.Next(maxY - (Pass.Height));
                    }
                    GenLocations[offset][k] = Pass;
                }
                offset = offset +2;
            }
            return offset;
        }

        //based on algorithm from snipplr.com/view/63587/shuffle--randomize-array/
        private void Randomize_Population(int gInd)
        {
            int indx, tscore = 0;
            Rectangle[] trect;
            for (int i = 0; i < gInd; i++)
            {
                indx = rand.Next(i, gInd);  //get next position
                    //swap elements
                tscore = genScores[i];          //swap scores array
                genScores[i] = genScores[indx];
                genScores[indx] = tscore;
                    //swap locations array
                trect = GenLocations[i];
                GenLocations[i] = GenLocations[indx];
                GenLocations[indx] = trect;
            }
        }

        private void onPaint(object sender, PaintEventArgs e)
        {
            Graphics g = this.CreateGraphics();
            g.DrawImage(img, 0, 0);
        }

        //when a tagphrase is clicked on, a messagebox shows the volume of that tag phrase
        private void ShowVolume(object sender, MouseEventArgs e)
        {
            Rectangle temp;
            for (int i = 0; i < tagphrase_total; i++)
            {
                temp = GenLocations[0][i];
                if (e.X > temp.Left && e.X < temp.Right && e.Y > temp.Top && e.Y < temp.Bottom)
                {
                    MessageBox.Show("Tag Phrase: " + arrData[i].TagPhrase + 
                        
                        " tweet volume: " + arrData[i].Volume.ToString());
                }
            }
        }

        //***************** OLD CODE *****************//
        //These are all old versions of functions that have been eliminated as I
        //progressed through development.

        //How I was placing text on screen when I was using text boxes... version 1
        public void ScreenPlacementBox()
        {
            TextBox[] boxDisplay = new TextBox[tagphrase_total];

            for (int i = 0; i < tagphrase_total; i++)
            {
                TextBox tempbox = new TextBox();
                //set font size
                tempbox.Font = new Font(tempbox.Font.FontFamily, arrData[i].Size);
                tempbox.Text = arrData[i].TagPhrase;    //populate with array of tagphrases
                tempbox.Enabled = false;                //disable the box
                tempbox.BorderStyle = System.Windows.Forms.BorderStyle.None;    //turn off border
                                                                                //resize the box
                Size tsize = TextRenderer.MeasureText(tempbox.Text, tempbox.Font);
                tempbox.Width = tsize.Width;
                tempbox.Height = tsize.Height;

                //generate a random location for the text box
                Random rand = new Random();
                int xplace = (rand.Next(maxX - tsize.Width));   //prevents from going off screen
                int yplace = (rand.Next(maxY - tsize.Height));  //by subtracting dimensions
                tempbox.Location = new Point(xplace, yplace);

                boxDisplay[i] = new TextBox();
                boxDisplay[i] = tempbox;
                this.Controls.Add(boxDisplay[i]);
            }
        }
        //How I was checking the overlap before... looking at position of rectangles
        //containing the text.  Probably don't need the rectangles array anymore.
        private int CheckOverlapOLD()
        {
            //This will be called to determine fitness of a particular configuration.
            //If there is overlap, there will be a substantially negative score returned
            //to the main program.  We do NOT want overlap.  If there is absolutely no
            //overlap, it will retun a high positive score.  The exact values that are
            //returned may be changed in the future, once I have the proximity check
            //written.  I want to optimize the values, too, so that I have an ideal
            //combination of no overlap and proximity.
            int returnvalue = 0;
            MessageBox.Show("rv est");
            int iLeft, jLeft = 0;
            int iRight, jRight = 0;
            int iTop, jTop = 0;
            int iBottom, jBottom = 0;

            //Look at each rectangle in the array of rectangles.  Compare it to the
            //rectangles that come after it.  For each overlap, subtract from the 
            //value to return.  
            for (int i = 0; i < tagphrase_total; i++)
            {
                MessageBox.Show("in i loop");
                //get coordinate values for initial rectangle
                iLeft = TagLocations[i].Left;       //x coordinates
                iRight = TagLocations[i].Right;
                iTop = TagLocations[i].Top;         //y coordinates
                iBottom = TagLocations[i].Bottom;
                for (int j = 1; j < tagphrase_total; j++)
                {
                    MessageBox.Show("j: " + j.ToString());
                    jLeft = TagLocations[j].Left;       //x coordinates
                    jRight = TagLocations[j].Right;
                    jTop = TagLocations[j].Top;         //y coordinates
                    jBottom = TagLocations[j].Bottom;
                    MessageBox.Show("j values done");
                    //is it "inside" the box coordinate? Check overlap for each corner
                    //the more overlap, the worse score... that's why "if" not "else if"
                    //(top & left j > top & left i AND top/left j < bottom/right i)
                    if (jTop > iTop && jLeft > iLeft)
                    {
                        if (jTop < iBottom && jLeft < iRight)
                            returnvalue += -2000;
                    }
                    //UR corner overlap: itop>jtop>ibot AND ileft>jright>iright
                    if (iTop > jTop && jTop > jBottom && iLeft > jRight && jRight > iRight)
                        returnvalue += -2000;
                    //LR corner overlap: itop>jbot>ibot AND ileft>jright>iright
                    if (jTop > jBottom && jBottom > iBottom && iLeft > jRight && jRight > iRight)
                        returnvalue += -2000;
                    //LL corner overlap: itop>jtop>ibot AND ileft>jleft>iright
                    if (iTop > jTop && jTop > iBottom && iLeft > jLeft && jLeft > iRight)
                        returnvalue += -2000;
                }
                MessageBox.Show(i.ToString() + " return value: " + returnvalue.ToString());
            }
            if (returnvalue == 0)   //no overlaps AT ALL; it still retains the initialized value
                returnvalue = 5000;
            return returnvalue;
        }
        //bring outliers closer to mutate to closer arrangement
        //Slowly decrease the offset.
        public Rectangle SnugUp(Rectangle Pass)
        {
            if (Pass.Top < 175)         //way up on the top
                Pass.Offset(0, 5);          //move down by 15
            if (Pass.Bottom > 400)      //way down on the bottom
                Pass.Offset(0, -5);         //move up by 15
            if (Pass.Left < 100)         //way to the left
                Pass.Offset(5, 0);          //move right by 7
            if (Pass.Right > 600)       //way to the right
                Pass.Offset(-5, 0);         //move to the left by 7
            return Pass;
        }
    }
}
